"""
PyPy-specific typing support for PEP 695 (Type Parameter Syntax).

This module provides TypeVar, ParamSpec, TypeVarTuple, TypeAliasType and Generic
implementations with support for lazy evaluation of bounds, constraints,
and type alias values.

These are moved here from typing.py to mirror CPython's approach where
these classes are implemented in C.

This module is intended as a hacky replacement for the relevant parts of typing.py
that were rewritten in C in the CPython support for PEP 695, to allow us to start
implementing PEP 695 support in PyPy without having to rewrite large parts of
the typing module in RPython.
A lot of this code must be rewritten in RPython later to establish proper
parity with CPython's implementation around introspection, immutability, etc.
"""

import sys

__all__ = [
    'TypeVar', 'ParamSpec', 'TypeVarTuple', 'TypeAliasType',
    'ParamSpecArgs', 'ParamSpecKwargs', 'Generic',
]


def _caller_module():
    # Return the __name__ of the module that called the __init__ invoking this
    # helper (frame 0 is _caller_module, frame 1 is the __init__, frame 2 is
    # its caller). Mirrors CPython's caller() in Objects/typevarobject.c, which
    # sets __module__ on explicitly constructed type variables to the defining
    # module (needed so they can be pickled by qualified name).
    try:
        f = sys._getframe(2)
    except ValueError:
        return None
    return f.f_globals.get('__name__')


class _Immutable:
    """Mixin that makes instances immutable: attribute assignment and
    deletion are blocked from outside this module, for a fixed set of
    "core" attributes given by _readonly_attrs.  This mirrors CPython,
    where these types are implemented in C with a handful of read-only
    struct members but otherwise allow arbitrary new instance attributes
    (e.g. typing_extensions patches 'has_default' onto TypeVar/ParamSpec
    instances on Python < 3.13).  _readonly_attrs of None means there are
    no such extra attributes and everything is blocked, matching types
    that CPython implements without a per-instance dict at all.  Internal
    code that legitimately needs to set up or lazily patch one of the
    readonly attributes must go through
    object.__setattr__/object.__delattr__ to bypass this.
    """
    __slots__ = ()

    _readonly_attrs = None

    def __setattr__(self, name, value):
        if self._readonly_attrs is None or name in self._readonly_attrs:
            raise TypeError(
                "cannot set %r attribute of immutable type %r" %
                (name, type(self).__name__))
        object.__setattr__(self, name, value)

    def __delattr__(self, name):
        if self._readonly_attrs is None or name in self._readonly_attrs:
            raise TypeError(
                "cannot delete %r attribute of immutable type %r" %
                (name, type(self).__name__))
        object.__delattr__(self, name)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class _PickleUsingNameMixin:
    """Mixin for types that can be pickled using their __name__."""

    def __reduce__(self):
        return self.__name__


class _BoundVarianceMixin:
    """Mixin giving __init__ bound and variance arguments.

    This is used by TypeVar and ParamSpec, which both employ the notions of
    a type 'bound' (restricting type arguments to be a subtype of some
    specified type) and type 'variance' (determining subtype relations between
    generic types).
    """
    def __init__(self, bound, covariant, contravariant, infer_variance):
        """Used to setup TypeVars and ParamSpec's bound, covariant,
        contravariant and infer_variance attributes.
        """
        if covariant and contravariant:
            raise ValueError("Bivariant types are not supported.")
        if infer_variance and (covariant or contravariant):
            raise ValueError("Variance cannot be specified with infer_variance.")
        object.__setattr__(self, '__covariant__', bool(covariant))
        object.__setattr__(self, '__contravariant__', bool(contravariant))
        object.__setattr__(self, '__infer_variance__', bool(infer_variance))
        if bound:
            from typing import _type_check
            bound = _type_check(bound, "Bound must be a type.")
        else:
            bound = None
        object.__setattr__(self, '__bound__', bound)

    def __or__(self, right):
        import typing
        return typing._make_union(self, right)

    def __ror__(self, left):
        import typing
        return typing._make_union(left, self)

    def __repr__(self):
        if self.__infer_variance__:
            prefix = ''
        elif self.__covariant__:
            prefix = '+'
        elif self.__contravariant__:
            prefix = '-'
        else:
            prefix = '~'
        return prefix + self.__name__


class _LazyEvaluator:
    __slots__ = ('_name',)

    def __set_name__(self, owner, name):
        assert name.startswith("__") and name.endswith("__")
        self._name = name[2:-2]

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = getattr(instance, f"__evaluate_{self._name}__")()
        object.__setattr__(instance, f"__{self._name}__", value)
        return value


class TypeVar(_Immutable, _PickleUsingNameMixin, _BoundVarianceMixin):
    """Type variable with support for lazy bound/constraints evaluation.

    Usage::

      T = TypeVar('T')  # Can be anything
      A = TypeVar('A', str, bytes)  # Must be str or bytes

    Type variables exist primarily for the benefit of static type
    checkers.  They serve as the parameters for generic types as well
    as for generic function definitions.
    """

    _readonly_attrs = frozenset((
        '__name__', '__bound__', '__covariant__', '__contravariant__',
        '__infer_variance__', '__constraints__',
    ))

    def __init__(self, name, *constraints, bound=None, covariant=False,
                 contravariant=False, infer_variance=False):
        object.__setattr__(self, '__name__', name)
        super().__init__(bound, covariant, contravariant, infer_variance)
        if constraints:
            if len(constraints) == 1:
                raise TypeError("A single constraint is not allowed")
            if bound is not None:
                raise TypeError("Constraints cannot be combined with bound=...")
            from typing import _type_check
            msg = "TypeVar(name, constraint, ...): constraints must be types."
            constraints = tuple(_type_check(t, msg) for t in constraints)
        else:
            constraints = ()
        object.__setattr__(self, '__constraints__', constraints)

        object.__setattr__(self, '__module__', _caller_module())

    __bound__ = _LazyEvaluator()
    __constraints__ = _LazyEvaluator()

    def __typing_subst__(self, arg):
        """Used for generic type substitution."""
        import typing
        return typing._typevar_subst(self, arg)

    def __mro_entries__(self, bases):
        raise TypeError("Cannot subclass an instance of TypeVar")

    def __init_subclass__(cls, **kwargs):
        raise TypeError("type 'typing.TypeVar' is not an acceptable base type")


class ParamSpec(_Immutable, _PickleUsingNameMixin, _BoundVarianceMixin):
    """Parameter specification variable.

    The preferred way to construct a parameter specification is via the
    dedicated syntax for generic functions, classes, and type aliases,
    where the use of '**' creates a parameter specification::

        type IntFunc[**P] = Callable[P, int]

    For compatibility with Python 3.11 and earlier, ParamSpec objects
    can also be created as follows::

        P = ParamSpec('P')

    Parameter specification variables exist primarily for the benefit of
    static type checkers.  They are used to forward the parameter types of
    one callable to another callable, a pattern commonly found in
    higher-order functions and decorators.  They are only valid when used
    in ``Concatenate``, or as the first argument to ``Callable``, or as
    parameters for user-defined Generics. See class Generic for more
    information on generic types.

    An example for annotating a decorator::

        def add_logging[**P, T](f: Callable[P, T]) -> Callable[P, T]:
            '''A type-safe decorator to add logging to a function.'''
            def inner(*args: P.args, **kwargs: P.kwargs) -> T:
                logging.info(f'{f.__name__} was called')
                return f(*args, **kwargs)
            return inner

        @add_logging
        def add_two(x: float, y: float) -> float:
            '''Add two numbers together.'''
            return x + y

    Parameter specification variables can be introspected. e.g.::

        >>> P = ParamSpec("P")
        >>> P.__name__
        'P'

    Note that only parameter specification variables defined in the global
    scope can be pickled.
    """

    _readonly_attrs = frozenset((
        '__name__', '__bound__', '__covariant__', '__contravariant__',
        '__infer_variance__', 'args', 'kwargs',
    ))

    def __init__(self, name, *, bound=None, covariant=False, contravariant=False, infer_variance=False):
        object.__setattr__(self, '__name__', name)
        super().__init__(bound, covariant, contravariant, infer_variance)

        object.__setattr__(self, '__module__', _caller_module())

        # Create args and kwargs attributes
        object.__setattr__(self, 'args', ParamSpecArgs(self))
        object.__setattr__(self, 'kwargs', ParamSpecKwargs(self))

    def __typing_subst__(self, arg):
        import typing
        return typing._paramspec_subst(self, arg)

    def __typing_prepare_subst__(self, alias, args):
        import typing
        return typing._paramspec_prepare_subst(self, alias, args)

    def __mro_entries__(self, bases):
        raise TypeError("Cannot subclass an instance of ParamSpec")

    def __init_subclass__(cls, **kwargs):
        raise TypeError("type 'typing.ParamSpec' is not an acceptable base type")


class ParamSpecArgs(_Immutable):
    """The args for a ParamSpec object.

    Given P = ParamSpec('P'), P.args is an instance of ParamSpecArgs.
    """
    __slots__ = ('__origin__',)

    def __init__(self, origin):
        object.__setattr__(self, '__origin__', origin)

    def __repr__(self):
        return f"{self.__origin__.__name__}.args"

    def __hash__(self):
        return hash((self.__origin__, "args"))

    def __eq__(self, other):
        if isinstance(other, ParamSpecArgs):
            return self.__origin__ == other.__origin__
        return NotImplemented

    def __mro_entries__(self, bases):
        raise TypeError("Cannot subclass an instance of ParamSpecArgs")

    def __init_subclass__(cls, **kwargs):
        raise TypeError("type 'typing.ParamSpecArgs' is not an acceptable base type")


class ParamSpecKwargs(_Immutable):
    """The kwargs for a ParamSpec object.

    Given P = ParamSpec('P'), P.kwargs is an instance of ParamSpecKwargs.
    """
    __slots__ = ('__origin__',)

    def __init__(self, origin):
        object.__setattr__(self, '__origin__', origin)

    def __repr__(self):
        return f"{self.__origin__.__name__}.kwargs"

    def __hash__(self):
        return hash((self.__origin__, "kwargs"))

    def __eq__(self, other):
        if isinstance(other, ParamSpecKwargs):
            return self.__origin__ == other.__origin__
        return NotImplemented

    def __mro_entries__(self, bases):
        raise TypeError("Cannot subclass an instance of ParamSpecKwargs")

    def __init_subclass__(cls, **kwargs):
        raise TypeError("type 'typing.ParamSpecKwargs' is not an acceptable base type")


class TypeVarTuple(_Immutable, _PickleUsingNameMixin):
    """Type variable tuple.

    Usage::

      Ts = TypeVarTuple('Ts')

    A TypeVarTuple is a placeholder for an *arbitrary* number of types.
    """

    _readonly_attrs = frozenset(('__name__',))

    __slots__ = ('__name__',)

    def __init__(self, name):
        object.__setattr__(self, '__name__', name)
        object.__setattr__(self, '__module__', _caller_module())

    def __repr__(self):
        return self.__name__

    def __iter__(self):
        from typing import Unpack
        yield Unpack[self]

    def __typing_subst__(self, arg):
        raise TypeError("Substitution of bare TypeVarTuple is not supported")

    def __typing_prepare_subst__(self, alias, args):
        import typing
        return typing._typevartuple_prepare_subst(self, alias, args)

    def __mro_entries__(self, bases):
        raise TypeError("Cannot subclass an instance of TypeVarTuple")

    def __init_subclass__(cls, **kwargs):
        raise TypeError("type 'typing.TypeVarTuple' is not an acceptable base type")


class TypeAliasType(_Immutable, _PickleUsingNameMixin):
    """Runtime representation of a type alias created with PEP 695 syntax.

    The __value__ is lazily evaluated - the evaluate_func is called
    only when __value__ is first accessed, then cached.

    Example::

        type Point = tuple[float, float]
        # Point is a TypeAliasType with __name__ = 'Point'
        # and __value__ = tuple[float, float]
    """

    def __init__(self, name, value, *, type_params=()):
        """Initialize a TypeAliasType.

        Args:
            name: The name of the type alias.
            value: The value of the type alias.
            type_params: The type parameters of the alias (for generic aliases).
        """
        object.__setattr__(self, '_name', name)
        object.__setattr__(self, '_type_params',
                           tuple(type_params) if type_params else ())
        object.__setattr__(self, '__value__', value)
        object.__setattr__(self, '__module__', _caller_module())

    @property
    def __name__(self):
        return self._name

    @property
    def __type_params__(self):
        return self._type_params

    @property
    def __parameters__(self):
        """Return the type parameters, unpacking any TypeVarTuples."""
        if not self._type_params:
            return ()
        result = []
        for param in self._type_params:
            if isinstance(param, TypeVarTuple):
                result.extend(param)
            else:
                result.append(param)
        return tuple(result)

    __value__ = _LazyEvaluator()

    def __repr__(self):
        return self._name

    def __getitem__(self, parameters):
        """Support generic type alias subscripting: Alias[T]."""
        if not self._type_params:
            raise TypeError("Only generic type aliases are subscriptable")
        from typing import _GenericAlias
        if not isinstance(parameters, tuple):
            parameters = (parameters,)
        return _GenericAlias(self, parameters)

    def __or__(self, other):
        """Support | for Union types."""
        from typing import Union
        return Union[self, other]

    def __ror__(self, other):
        """Support | for Union types (reverse)."""
        from typing import Union
        return Union[other, self]


# Factory functions for the compiler
# These are called from generated bytecode to create type parameters
# with lazy evaluation support.

def _make_typevar(name):
    t = TypeVar(name, infer_variance=True)
    object.__setattr__(t, '__module__', 'typing')
    return t


def _make_typevar_with_bound(name, evaluate_bound):
    t = TypeVar(name, infer_variance=True)
    object.__delattr__(t, '__bound__')
    object.__setattr__(t, '__evaluate_bound__', evaluate_bound)
    object.__setattr__(t, '__module__', 'typing')
    return t


def _make_typevar_with_constraints(name, evaluate_constraints):
    t = TypeVar(name, infer_variance=True)
    object.__delattr__(t, '__constraints__')
    object.__setattr__(t, '__evaluate_constraints__', evaluate_constraints)
    object.__setattr__(t, '__module__', 'typing')
    return t


def _make_paramspec(name):
    t = ParamSpec(name, infer_variance=True)
    object.__setattr__(t, '__module__', 'typing')
    return t


def _make_typevartuple(name):
    t = TypeVarTuple(name)
    object.__setattr__(t, '__module__', 'typing')
    return t


def _make_typealiastype(name, evaluate_value, type_params):
    t = TypeAliasType(name, None, type_params=type_params)
    object.__delattr__(t, '__value__')
    object.__setattr__(t, '__evaluate_value__', evaluate_value)
    object.__setattr__(t, '__module__', getattr(evaluate_value, '__module__', None))
    return t


class Generic:
    """Abstract base class for generic types.

    A generic type is typically declared by inheriting from
    this class parameterized with one or more type variables.
    For example, a generic mapping type might be defined as::

      class Mapping(Generic[KT, VT]):
          def __getitem__(self, key: KT) -> VT:
              ...
          # Etc.

    This class can then be used as follows::

      def lookup_name(mapping: Mapping[KT, VT], key: KT, default: VT) -> VT:
          try:
              return mapping[key]
          except KeyError:
              return default
    """
    __slots__ = ()
    _is_protocol = False

    def __class_getitem__(cls, params):
        import typing
        # PEP 695: the compiler builds the implicit Generic[...] base from a
        # tuple of the class's bare type params (Generic[(T, Ts, P)], always
        # a genuine tuple, built by BUILD_TUPLE). CPython's
        # _Py_subscript_generic intrinsic unpacks any TypeVarTuple (Ts -> *Ts)
        # before the subscript, which _generic_class_getitem then requires (a
        # bare TypeVarTuple is not "type-var-like"). PyPy routes that
        # intrinsic through Generic.__class_getitem__ itself, so mirror the
        # unpacking here -- but only for that specific cls-is-Generic, tuple
        # call shape. An ordinary bare TypeVarTuple subscript of Generic
        # itself (`Generic[Ts]`, no comma) is user code, not the compiler,
        # and must stay invalid; likewise subscripting an already-defined
        # subclass (`SomeGeneric[Ts]`) must leave a bare Ts as an
        # unsubstitutable placeholder rather than silently unpacking it.
        if cls is Generic and isinstance(params, tuple):
            params = tuple(
                typing.Unpack[p] if isinstance(p, typing.TypeVarTuple) else p
                for p in params)
        return typing._generic_class_getitem(cls, params)

    def __init_subclass__(cls, *args, **kwargs):
        import typing
        return typing._generic_init_subclass(cls, *args, **kwargs)


# make the __module__ match pickling by their public name.
for _cls in (TypeVar, ParamSpec, TypeVarTuple, TypeAliasType, Generic,
             ParamSpecArgs, ParamSpecKwargs):
    _cls.__module__ = 'typing'
del _cls



"""A small, distutils-free compiler driver for building cffi extension modules.

On Python >= 3.12 the stdlib ``distutils`` is gone and cffi normally requires
``setuptools`` to compile out-of-line modules (see
``cffi/_shimmed_dist_utils.py``).  During PyPy's own build that creates a
chicken-and-egg problem: ``build_cffi_imports`` needs to build ``_ssl`` (used by
hashlib/ssl) *before* ``ensurepip`` can install ``setuptools``, but ``ensurepip``
itself needs ``_ssl``.

This module breaks the cycle by invoking the C compiler and linker directly,
the same way distutils' ``UnixCCompiler``/``MSVCCompiler`` do, but with no
distutils/setuptools dependency.  The toolchain is discovered in this order:

1. the ``PYPY_CC`` / ``PYPY_CC_LINK`` / ``PYPY_CC_KIND`` environment variables,
   which PyPy's translation driver sets from ``rpython.translator.platform``
   (this is how the Windows MSVC environment is handed down -- the host already
   located ``cl.exe``/``link.exe`` and their ``INCLUDE``/``LIB`` env);
2. ``sysconfig`` config vars (``CC``, ``LDSHARED``, ...), which a normal
   installed PyPy populates from ``_sysconfigdata`` -- this keeps the
   user-facing ``pypy build_cffi_imports.py`` rebuild working standalone on
   POSIX;
3. nothing else -- if neither is available a clear error is raised.
"""
import os
import shlex
import subprocess
import sys
import sysconfig


class Extension(object):
    """A drop-in stand-in for ``distutils.core.Extension``.

    Only the attributes cffi actually passes through ``get_extension`` are
    honoured; any other distutils keyword (``depends``, ...) is accepted and
    ignored so existing build scripts keep working.
    """
    def __init__(self, name, sources, include_dirs=None, define_macros=None,
                 undef_macros=None, library_dirs=None, libraries=None,
                 extra_compile_args=None, extra_link_args=None,
                 extra_objects=None, **ignored):
        self.name = name
        self.sources = list(sources)
        self.include_dirs = list(include_dirs or [])
        self.define_macros = list(define_macros or [])
        self.undef_macros = list(undef_macros or [])
        self.library_dirs = list(library_dirs or [])
        self.libraries = list(libraries or [])
        self.extra_compile_args = list(extra_compile_args or [])
        self.extra_link_args = list(extra_link_args or [])
        self.extra_objects = list(extra_objects or [])


def _split(cmd):
    if not cmd:
        return []
    if isinstance(cmd, (list, tuple)):
        return list(cmd)
    return shlex.split(cmd, posix=(os.name != 'nt'))


def _kind():
    kind = os.environ.get('PYPY_CC_KIND')
    if kind:
        return kind
    return 'msvc' if sys.platform == 'win32' else 'unix'


def _cc():
    return _split(os.environ.get('PYPY_CC') or
                  sysconfig.get_config_var('CC'))


def _linker(default_cc):
    link = _split(os.environ.get('PYPY_CC_LINK') or
                  sysconfig.get_config_var('LDSHARED'))
    if not link:
        link = list(default_cc)
    return link


def _ext_suffix():
    return (sysconfig.get_config_var('EXT_SUFFIX') or
            sysconfig.get_config_var('SO') or
            ('.pyd' if sys.platform == 'win32' else '.so'))


def _output_filename(modname):
    # mirror distutils.command.build_ext.get_ext_filename: a dotted name maps
    # to a path, with the platform extension suffix appended.
    parts = modname.split('.')
    return os.path.join(*parts) + _ext_suffix()


def _ensure_parent(path):
    parent = os.path.dirname(path)
    if parent and not os.path.isdir(parent):
        os.makedirs(parent)


def _run(args, verbose):
    if verbose:
        print(' '.join(args))
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    out, _ = proc.communicate()
    if proc.returncode != 0:
        from cffi.error import VerificationError
        if not isinstance(out, str):
            out = out.decode('utf-8', 'replace')
        raise VerificationError(
            '%s\ncommand failed (status %d): %s' %
            (out, proc.returncode, ' '.join(args)))


def _macros(define_macros, undef_macros, prefix):
    args = []
    for name, value in define_macros:
        if value is None:
            args.append('%s%s' % (prefix, name))
        else:
            args.append('%s%s=%s' % (prefix, name, value))
    for name in undef_macros:
        args.append('-U' + name)
    return args


def _build_unix(ext, verbose):
    cc = _cc() or ['cc']
    link = _linker(cc)
    # make sure we actually produce a shared object
    if not (set(link) & set(['-shared', '-bundle', '-dynamiclib'])):
        link = link + ['-shared']

    # honour the standard compiler env vars the way distutils' UnixCCompiler
    # does -- create_cffi_import_libraries injects the PyPy INCLUDEPY into
    # CFLAGS and (for --embed-dependencies) the bundled openssl dirs into
    # CPPFLAGS/LDFLAGS.
    cflags = _split(os.environ.get('CFLAGS') or
                    sysconfig.get_config_var('CFLAGS'))
    cppflags = _split(os.environ.get('CPPFLAGS'))
    ldflags = _split(os.environ.get('LDFLAGS'))
    ccshared = _split(sysconfig.get_config_var('CCSHARED'))
    if not ccshared:
        ccshared = ['-fPIC']
    includepy = sysconfig.get_config_var('INCLUDEPY')

    objects = []
    for src in ext.sources:
        obj = os.path.splitext(os.path.basename(src))[0] + '.o'
        args = list(cc) + cflags + cppflags + ccshared
        if includepy:
            args.append('-I' + includepy)
        args += ['-I' + d for d in ext.include_dirs]
        args += _macros(ext.define_macros, ext.undef_macros, '-D')
        args += ext.extra_compile_args
        args += ['-c', src, '-o', obj]
        _run(args, verbose)
        objects.append(obj)

    out = _output_filename(ext.name)
    _ensure_parent(out)
    args = list(link) + ldflags + objects + ext.extra_objects
    args += ['-L' + d for d in ext.library_dirs]
    args += ['-l' + l for l in ext.libraries]
    args += ext.extra_link_args
    args += ['-o', out]
    _run(args, verbose)
    return out


def _msvc_library_dirs(ext):
    # distutils' build_ext used to append these for every win32 extension
    # (that's where e.g. python312.lib, pulled in via pyconfig.h's #pragma
    # comment(lib, ...), gets found); replicate it since distutils is gone.
    library_dirs = list(ext.library_dirs)
    library_dirs.append(os.path.join(sys.exec_prefix, 'libs'))
    if sys.base_exec_prefix != sys.exec_prefix:  # e.g. running in a venv
        library_dirs.append(os.path.join(sys.base_exec_prefix, 'libs'))
    return library_dirs


def _build_msvc(ext, verbose):
    # NB: validated on POSIX; the MSVC command lines mirror distutils'
    # MSVCCompiler and rely on the translation driver having exported the
    # vcvars environment (INCLUDE/LIB/PATH) so cl.exe/link.exe are usable.
    cc = _cc() or ['cl.exe']
    link = _split(os.environ.get('PYPY_CC_LINK')) or ['link.exe']
    includepy = sysconfig.get_config_var('INCLUDEPY')

    objects = []
    for src in ext.sources:
        obj = os.path.splitext(os.path.basename(src))[0] + '.obj'
        args = list(cc) + ['/nologo', '/O2', '/MD', '/c']
        if includepy:
            args.append('/I' + includepy)
        args += ['/I' + d for d in ext.include_dirs]
        args += _macros(ext.define_macros, ext.undef_macros, '/D')
        args += ext.extra_compile_args
        args += [src, '/Fo' + obj]
        _run(args, verbose)
        objects.append(obj)

    out = _output_filename(ext.name)
    _ensure_parent(out)
    # the C extension must export its module init function
    init = 'PyInit_' + ext.name.split('.')[-1]
    args = list(link) + ['/nologo', '/DLL', '/EXPORT:' + init]
    args += ['/OUT:' + out] + objects + ext.extra_objects
    args += ['/LIBPATH:' + d for d in _msvc_library_dirs(ext)]
    args += [l if l.lower().endswith('.lib') else l + '.lib'
             for l in ext.libraries]
    args += ext.extra_link_args
    _run(args, verbose)
    return out


def build(ext, tmpdir='.', compiler_verbose=0, debug=None):
    """Compile and link ``ext`` (an :class:`Extension`) in ``tmpdir``.

    Returns the path to the produced shared object (relative to ``tmpdir``,
    which the caller has already ``chdir``-ed into, matching
    ``cffi.ffiplatform``).
    """
    if _kind() == 'msvc':
        return _build_msvc(ext, compiler_verbose)
    return _build_unix(ext, compiler_verbose)


def compile_shared(csource, modulename, output_dir, include_dirs=None,
                   define_macros=None, libraries=None, library_dirs=None,
                   extra_compile_args=None, extra_link_args=None,
                   compiler_verbose=0):
    """Compile one or more C sources into an extension module, distutils-free.

    A sibling of :func:`cffi.ffiplatform.compile`: both sit on
    :class:`Extension`/:func:`build`, but this one packages the build the
    on-demand C test modules (``_testcapi``, ``_ctypes_test``, ...) used to do
    through ``distutils.ccompiler``.  ``csource`` is a single source path or a
    list of them; they are compiled and linked into ``modulename`` inside
    ``output_dir`` and the absolute path of the produced shared object is
    returned.  The PyPy header dir (INCLUDEPY) and, on MSVC, the ``PyInit_``
    export are added by :func:`build` itself.
    """
    if isinstance(csource, str):
        sources = [csource]
    else:
        sources = list(csource)
    ext = Extension(modulename, sources, include_dirs=include_dirs,
                    define_macros=define_macros, libraries=libraries,
                    library_dirs=library_dirs,
                    extra_compile_args=extra_compile_args,
                    extra_link_args=extra_link_args)
    oldcwd = os.getcwd()
    os.chdir(output_dir)
    try:
        out = build(ext, output_dir, compiler_verbose)
    finally:
        os.chdir(oldcwd)
    return os.path.join(output_dir, out)



from importlib.util import spec_from_file_location, module_from_spec
import os

try:
    import cpyext
except ImportError:
    pass   # no 'cpyext', but we still have to define e.g. awaitType
else:
    import _pypy_testcapi
    cfile = '_testcapimodule.c'
    csources = [cfile,
                '_testcapi/abstract.c',
                '_testcapi/buffer.c',
                '_testcapi/bytearray.c',
                #'_testcapi/bytes.c',
                #'_testcapi/code.c',
                '_testcapi/codec.c',
                #'_testcapi/complex.c',
                '_testcapi/datetime.c',
                #'_testcapi/dict.c',
                '_testcapi/docstring.c',
                '_testcapi/eval.c',
                '_testcapi/exceptions.c',
                '_testcapi/file.c',
                #'_testcapi/float.c',
                #'_testcapi/getargs.c',
                #'_testcapi/heaptype.c',
                '_testcapi/heaptype_relative.c',
                '_testcapi/immortal.c',
                #'_testcapi/import.c',
                '_testcapi/list.c',
                #'_testcapi/long.c',
                #'_testcapi/mem.c',
                '_testcapi/numbers.c',
                #'_testcapi/pyos.c',
                '_testcapi/pytime.c',
                '_testcapi/run.c',
                '_testcapi/set.c',
                '_testcapi/structmember.c',
                #'_testcapi/sys.c',
                '_testcapi/tuple.c',
                #'_testcapi/unicode.c',
                '_testcapi/vectorcall.c',
                '_testcapi/vectorcall_limited.c',
                #'_testcapi/watchers.c',
               ]
    thisdir = os.path.dirname(__file__)
    output_dir = _pypy_testcapi.get_hashed_dir(os.path.join(thisdir, cfile))
    modfile = '_testcapi' + _pypy_testcapi._get_c_extension_suffix()
    spec = spec_from_file_location('_testcapi',
                                   os.path.join(thisdir, modfile))
    try:
        module_from_spec(spec)
    except ImportError:
        if os.name == 'nt':
            # hack around finding compilers on win32
            try:
                import setuptools
            except ImportError:
                pass
        mod = _pypy_testcapi.compile_shared(csources, '_testcapi', thisdir)

class awaitType:
    def __init__(self, iterator):
        self._iterator = iterator
    def __await__(self):
        return self._iterator

# the hacks above have replaced this module with another, so we need
# to push the extra names into this other module too...
import _testcapi
_testcapi.awaitType = awaitType



"""
This will create the _ctypes_test c-extension module. Unlike _testcapi, the
extension cannot be wrapped with a _ctypes_test.py module since test.importlib
explicitly does a c-extension import
"""
import importlib.machinery
import importlib.util
import os
import sys

try:
    import cpyext
except ImportError:
    raise RuntimeError("must have cpyext")
import _pypy_testcapi
cfile = '_ctypes_test.c'
thisdir = os.path.dirname(__file__)
output_dir = _pypy_testcapi.get_hashed_dir(os.path.join(thisdir, cfile))
try:
    import _ctypes
except ImportError:
    pass    # obscure condition of _ctypes_test.py being imported by py.test
else:
    _ctypes.PyObj_FromPtr = None
    del _ctypes
    try:
        # Look for an already-compiled _ctypes_test extension in thisdir
        # (replaces imp.find_module/imp.load_module, removed in 3.12).
        filename = None
        for suffix in importlib.machinery.EXTENSION_SUFFIXES:
            candidate = os.path.join(thisdir, '_ctypes_test' + suffix)
            if os.path.exists(candidate):
                filename = candidate
                break
        if filename is None:
            raise ImportError('_ctypes_test')
        spec = importlib.util.spec_from_file_location('_ctypes_test', filename)
        module = importlib.util.module_from_spec(spec)
        sys.modules['_ctypes_test'] = module
        spec.loader.exec_module(module)
    except ImportError:
        _pypy_testcapi.compile_shared('_ctypes_test.c', '_ctypes_test', thisdir)



"""_typing: re-export PyPy's typing primitives so that lib-python/3/typing.py
finds the same TypeVar/ParamSpec/Generic classes that PyPy's PEP 695 runtime
uses when creating type parameters from the 'class Foo[T]' syntax.

If _typing and _pypy_typing exposed different TypeVar classes, isinstance
checks in typing._is_typevar_like would fail for PEP 695 type variables.
"""
from _pypy_typing import (   # noqa: F401
    TypeVar,
    ParamSpec,
    TypeVarTuple,
    ParamSpecArgs,
    ParamSpecKwargs,
    TypeAliasType,
    Generic,
)


def _idfunc(_, x):
    return x



import os, sys

import tempfile, binascii
import importlib.machinery
from importlib.util import spec_from_file_location, module_from_spec


def _get_hashed_filename(cfile):
    with open(cfile,'r') as fid:
        content = fid.read()
    # from cffi's Verifier()
    key = '\x00'.join([sys.version[:3], content])
    # change the key to force recompilation
    key += '2017-11-21'
    if sys.version_info >= (3,):
        key = key.encode('utf-8')
    k1 = hex(binascii.crc32(key[0::2]) & 0xffffffff)
    k1 = k1.lstrip('0x').rstrip('L')
    k2 = hex(binascii.crc32(key[1::2]) & 0xffffffff)
    k2 = k2.lstrip('0').rstrip('L')
    try:
        username = os.environ['USER']           #linux, et al
    except KeyError:
        try:
            username = os.environ['USERNAME']   #windows
        except KeyError:
            username = os.getuid()
    return tempfile.gettempdir() + os.path.sep + 'testcapi_%s_%s%s' % (
        username, k1, k2)

def get_hashed_dir(cfile):
    hashed_fn = _get_hashed_filename(cfile)
    try:
        with open(hashed_fn) as f:
            dirname = f.read(1024)
    except IOError:
        dirname = ''
    tmpdir = tempfile.gettempdir()
    if (not dirname or '/' in dirname or '\\' in dirname or '\x00' in dirname
            or not os.path.isdir(os.path.join(tmpdir, dirname))):
        dirname = binascii.hexlify(os.urandom(8))
        if not isinstance(dirname, str):    # Python 3
            dirname = dirname.decode('ascii')
        dirname = 'testcapi_' + dirname
    output_dir = os.path.join(tmpdir, dirname)
    try:
        os.mkdir(output_dir)
    except OSError:
        pass
    return output_dir


def _get_c_extension_suffix():
    suffixes = importlib.machinery.EXTENSION_SUFFIXES
    return suffixes[0] if suffixes else None


def _compile_shared_distutils(source_paths, modulename, output_dir):
    thisdir = os.path.dirname(__file__)
    from distutils.ccompiler import new_compiler
    from distutils import log, sysconfig
    log.set_verbosity(3)

    compiler = new_compiler()
    compiler.output_dir = output_dir
    # Compile .c files
    include_dir = sysconfig.get_config_var('INCLUDEPY')
    if sys.platform == 'win32':
        ccflags = ['-D_CRT_SECURE_NO_WARNINGS']
    else:
        ccflags = ['-fPIC', '-Wimplicit-function-declaration', '-O0', '-g3']
    sysconfig.customize_compiler(compiler)
    object_filenames = compiler.compile(source_paths,
                           include_dirs=[include_dir],
                           extra_preargs=ccflags,
                          )

    # set link options
    output_filename = modulename + _get_c_extension_suffix()
    if sys.platform == 'win32':
        libname = 'python{0[0]}{0[1]}'.format(sys.version_info)
        library = os.path.join(thisdir, '..', 'libs', libname)
        if not os.path.exists(library + '.lib'):
            # For a local translation or nightly build
            library = os.path.join(thisdir, '..', 'pypy', 'goal', libname)
        assert os.path.exists(library + '.lib'), 'Could not find import library "%s"' % library
        libraries = [library, 'oleaut32']
        extra_ldargs = ['/MANIFEST',  # needed for VC10
                        '/EXPORT:PyInit_' + modulename]
    else:
        libraries = []
        extra_ldargs = []

    # link the dynamic library
    compiler.link_shared_object(
        object_filenames,
        output_filename,
        libraries=libraries,
        extra_preargs=extra_ldargs)
    return os.path.join(output_dir, output_filename)


def _compile_shared_ccdriver(source_paths, modulename, output_dir):
    # distutils-free fallback for Python >= 3.12 (no distutils/setuptools).
    # Uses PyPy's own compiler driver; INCLUDEPY and the PyInit_ export are
    # added by the driver.
    from pypy_tools._cffi_compile import compile_shared as _cc_compile_shared
    thisdir = os.path.dirname(__file__)
    if sys.platform == 'win32':
        extra_compile_args = ['-D_CRT_SECURE_NO_WARNINGS']
        libname = 'python{0[0]}{0[1]}'.format(sys.version_info)
        library = os.path.join(thisdir, '..', 'libs', libname)
        if not os.path.exists(library + '.lib'):
            # For a local translation or nightly build
            library = os.path.join(thisdir, '..', 'pypy', 'goal', libname)
        assert os.path.exists(library + '.lib'), 'Could not find import library "%s"' % library
        libraries = [library, 'oleaut32']
        extra_link_args = ['/MANIFEST']  # needed for VC10
    else:
        extra_compile_args = ['-Wimplicit-function-declaration', '-O0', '-g3']
        libraries = None
        extra_link_args = None
    return _cc_compile_shared(source_paths, modulename, output_dir,
                              libraries=libraries,
                              extra_compile_args=extra_compile_args,
                              extra_link_args=extra_link_args)


def compile_shared(csource, modulename, output_dir):
    """Compile csource (a single filename or a list of them) into an
    extension module, and import it.
    """
    thisdir = os.path.dirname(__file__)
    assert output_dir is not None

    if isinstance(csource, str):
        csources = [csource]
    else:
        csources = list(csource)
    source_paths = [os.path.join(thisdir, c) for c in csources]
    try:
        output_filename = _compile_shared_distutils(
            source_paths, modulename, output_dir)
    except ImportError:
        output_filename = _compile_shared_ccdriver(
            source_paths, modulename, output_dir)

    # Now import the newly created library, it will replace the original
    # module in sys.modules
    spec = spec_from_file_location(modulename, output_filename)
    mod = module_from_spec(spec)

    # If everything went fine up to now, write the name of this new
    # directory to 'hashed_fn', for future processes (and to avoid a
    # growing number of temporary directories that are not completely
    # obvious to clean up on Windows)
    hashed_fn = _get_hashed_filename(source_paths[0])
    try:
        with open(hashed_fn, 'w') as f:
            f.write(os.path.basename(output_dir))
    except IOError:
        pass
    return mod



"""
Pure-Python implementation of _tokenize.TokenizerIter for PyPy 3.12.
Provides the same interface as CPython's Python/Python-tokenize.c.
"""

import re
import functools
import itertools as _itertools
from token import (
    ENDMARKER, NAME, NUMBER, STRING, NEWLINE, INDENT, DEDENT,
    OP, AWAIT, ASYNC, FSTRING_START, FSTRING_MIDDLE, FSTRING_END,
    COMMENT, NL, ERRORTOKEN, ENCODING,
    EXACT_TOKEN_TYPES,
)

# -- Regex patterns (adapted from tokenize.py) --------------------------------

def _group(*choices): return '(' + '|'.join(choices) + ')'
def _any(*choices):   return _group(*choices) + '*'
def _maybe(*choices): return _group(*choices) + '?'

Whitespace = r'[ \f\t]*'
Comment    = r'#[^\r\n]*'
Ignore     = Whitespace + _any(r'\\\r?\n' + Whitespace) + _maybe(Comment)
Name       = r'\w+'

Hexnumber  = r'0[xX](?:_?[0-9a-fA-F])+'
Binnumber  = r'0[bB](?:_?[01])+'
Octnumber  = r'0[oO](?:_?[0-7])+'
Decnumber  = r'(?:0(?:_?[0-9])*|[1-9](?:_?[0-9])*)'
Intnumber  = _group(Hexnumber, Binnumber, Octnumber, Decnumber)
Exponent   = r'[eE][-+]?[0-9](?:_?[0-9])*'
Pointfloat = _group(r'[0-9](?:_?[0-9])*\.(?:[0-9](?:_?[0-9])*)?',
                    r'\.[0-9](?:_?[0-9])*') + _maybe(Exponent)
Expfloat   = r'[0-9](?:_?[0-9])*' + Exponent
Floatnumber = _group(Pointfloat, Expfloat)
Imagnumber  = _group(r'[0-9](?:_?[0-9])*[jJ]', Floatnumber + r'[jJ]')
Number      = _group(Imagnumber, Floatnumber, Intnumber)


@functools.lru_cache(maxsize=None)
def _compile(expr):
    return re.compile(expr, re.UNICODE)


def _all_string_prefixes():
    _valid = ['b', 'r', 'u', 'f', 'br', 'fr']
    result = {''}
    for prefix in _valid:
        for t in _itertools.permutations(prefix):
            for u in _itertools.product(*[(c, c.upper()) for c in t]):
                result.add(''.join(u))
    return result


_ALL_STRING_PREFIXES = _all_string_prefixes()
StringPrefix = _group(*_ALL_STRING_PREFIXES)

Single  = r"[^'\\]*(?:\\.[^'\\]*)*'"
Double  = r'[^"\\]*(?:\\.[^"\\]*)*"'
Single3 = r"[^'\\]*(?:(?:\\.|'(?!''))[^'\\]*)*'''"
Double3 = r'[^"\\]*(?:(?:\\.|"(?!""))[^"\\]*)*"""'
Triple  = _group(StringPrefix + "'''", StringPrefix + '"""')
String  = _group(StringPrefix + r"'[^\n'\\]*(?:\\.[^\n'\\]*)*'",
                 StringPrefix + r'"[^\n"\\]*(?:\\.[^\n"\\]*)*"')

Special  = _group(*map(re.escape, sorted(EXACT_TOKEN_TYPES, reverse=True)))
Funny    = _group(r'\r?\n', Special)
ContStr  = _group(StringPrefix + r"'[^\n'\\]*(?:\\.[^\n'\\]*)*" +
                  _group("'", r'\\\r?\n'),
                  StringPrefix + r'"[^\n"\\]*(?:\\.[^\n"\\]*)*' +
                  _group('"', r'\\\r?\n'))
PseudoExtras = _group(r'\\\r?\n|\Z', Comment, Triple)
PseudoToken  = Whitespace + _group(PseudoExtras, Number, Funny, ContStr, Name)

endpats = {}
for _pfx in _ALL_STRING_PREFIXES:
    endpats[_pfx + "'"]   = Single
    endpats[_pfx + '"']   = Double
    endpats[_pfx + "'''"] = Single3
    endpats[_pfx + '"""'] = Double3
del _pfx

single_quoted = set()
triple_quoted = set()
for _t in _ALL_STRING_PREFIXES:
    for _u in (_t + '"', _t + "'"):
        single_quoted.add(_u)
    for _u in (_t + '"""', _t + "'''"):
        triple_quoted.add(_u)
del _t, _u

tabsize = 8


# -- F-string detection -------------------------------------------------------

def _fstring_info(s):
    """Return (prefix, quote) if s starts with an f-string, else (None, None)."""
    i = 0
    while i < len(s) and s[i] in 'fFrRbBuU':
        i += 1
    prefix = s[:i]
    if 'f' not in prefix.lower():
        return None, None
    rest = s[i:]
    if rest.startswith('"""') or rest.startswith("'''"):
        return prefix, rest[:3]
    if rest and rest[0] in '"\'':
        return prefix, rest[0]
    return None, None


# -- F-string body scanner ----------------------------------------------------

class _FStr:
    """Cursor for scanning text in a (possibly nested) f-string body."""

    __slots__ = ('body', 'pos', 'row', 'col', 'source_lines', 'n')

    def __init__(self, body, start_row, start_col, source_lines):
        self.body = body
        self.pos = 0
        self.row = start_row
        self.col = start_col
        self.source_lines = source_lines
        self.n = len(body)

    def src_line(self, r=None):
        if r is None:
            r = self.row
        sl = self.source_lines
        if sl and 1 <= r <= len(sl):
            return sl[r - 1]
        return ''

    def peek(self, offset=0):
        i = self.pos + offset
        if i < self.n:
            return self.body[i]
        return ''

    def at_end(self):
        return self.pos >= self.n

    def advance(self):
        ch = self.body[self.pos]
        self.pos += 1
        if ch == '\n':
            self.row += 1
            self.col = 0
        else:
            self.col += 1
        return ch

    def advance_by(self, text):
        """Advance cursor by the characters in text."""
        for ch in text:
            self.pos += 1
            if ch == '\n':
                self.row += 1
                self.col = 0
            else:
                self.col += 1

    def match_pseudo(self):
        return _compile(PseudoToken).match(self.body, self.pos)


def _scan_body(fs):
    """Yield FSTRING_MIDDLE and expression tokens for the f-string body."""
    seg_row, seg_col = fs.row, fs.col
    seg = []

    while not fs.at_end():
        ch = fs.peek()

        if ch == '{':
            if fs.peek(1) == '{':
                # {{ collapses to a single { that ENDS this segment; the end
                # column is the first brace + 1 and the second brace column is
                # absorbed (a new segment starts after both).  Matches CPython.
                seg.append('{')
                yield (FSTRING_MIDDLE, ''.join(seg),
                       (seg_row, seg_col), (fs.row, fs.col + 1),
                       fs.src_line(seg_row))
                fs.advance()
                fs.advance()
                seg = []
                seg_row, seg_col = fs.row, fs.col
            else:
                # Start of expression
                if seg:
                    yield (FSTRING_MIDDLE, ''.join(seg),
                           (seg_row, seg_col), (fs.row, fs.col),
                           fs.src_line(seg_row))
                    seg = []
                open_r, open_c = fs.row, fs.col
                fs.advance()
                yield (OP, '{', (open_r, open_c), (fs.row, fs.col),
                       fs.src_line(open_r))
                yield from _scan_expr(fs)
                seg_row, seg_col = fs.row, fs.col
        elif ch == '}':
            if fs.peek(1) == '}':
                # }} collapses to a single } that ENDS this segment (see above)
                seg.append('}')
                yield (FSTRING_MIDDLE, ''.join(seg),
                       (seg_row, seg_col), (fs.row, fs.col + 1),
                       fs.src_line(seg_row))
                fs.advance()
                fs.advance()
                seg = []
                seg_row, seg_col = fs.row, fs.col
            else:
                # Stray } (shouldn't happen in valid body)
                seg.append(ch)
                fs.advance()
        else:
            seg.append(ch)
            fs.advance()

    if seg:
        yield (FSTRING_MIDDLE, ''.join(seg),
               (seg_row, seg_col), (fs.row, fs.col),
               fs.src_line(seg_row))


def _scan_expr(fs):
    """
    Yield tokens for the f-string expression from current pos to matching }.
    Handles conversion (!r/!s/!a), format spec (:), debug (=), nested strings.
    Exits after consuming the closing }.
    """
    paren_depth = 0

    while not fs.at_end():
        # Newlines inside expressions become NL tokens
        if fs.peek() == '\n':
            nl_r, nl_c = fs.row, fs.col
            fs.advance()
            yield (NL, '\n', (nl_r, nl_c), (nl_r, nl_c + 1),
                   fs.src_line(nl_r))
            continue

        # Try pseudo-token match (handles whitespace prefix automatically)
        pm = fs.match_pseudo()
        if not pm:
            err_r, err_c = fs.row, fs.col
            ch = fs.advance()
            yield (ERRORTOKEN, ch, (err_r, err_c), (fs.row, fs.col),
                   fs.src_line(err_r))
            continue

        tok_start, tok_end = pm.span(1)
        if tok_start == tok_end:
            # Empty match (e.g. \Z); just advance one char
            if not fs.at_end():
                fs.advance()
            break

        # Advance past any leading whitespace (pos..tok_start)
        while fs.pos < tok_start:
            fs.advance()

        raw_tok = fs.body[tok_start:tok_end]
        initial = raw_tok[0]

        # --- Special cases at depth 0 ---

        # Closing brace at depth 0: end of expression
        if initial == '}' and paren_depth == 0:
            close_r, close_c = fs.row, fs.col
            fs.advance()
            yield (OP, '}', (close_r, close_c), (fs.row, fs.col),
                   fs.src_line(close_r))
            return

        # Colon at depth 0: format spec follows
        if initial == ':' and raw_tok == ':' and paren_depth == 0:
            col_r, col_c = fs.row, fs.col
            fs.advance()
            yield (OP, ':', (col_r, col_c), (fs.row, fs.col),
                   fs.src_line(col_r))
            yield from _scan_format_spec(fs)
            return

        # Conversion specifier !r/!s/!a at depth 0
        if initial == '!' and paren_depth == 0:
            nch = fs.peek(1)
            if nch in ('r', 's', 'a'):
                nnch = fs.peek(2)
                if nnch in ('}', ':', '\n', ''):
                    bang_r, bang_c = fs.row, fs.col
                    fs.advance()
                    yield (OP, '!', (bang_r, bang_c), (fs.row, fs.col),
                           fs.src_line(bang_r))
                    name_r, name_c = fs.row, fs.col
                    fs.advance()
                    yield (NAME, nch, (name_r, name_c), (fs.row, fs.col),
                           fs.src_line(name_r))
                    continue

        # Debug expression = at depth 0 (not ==, !=, <=, >=, :=)
        if (initial == '=' and raw_tok == '=' and paren_depth == 0):
            eq_r, eq_c = fs.row, fs.col
            fs.advance()
            yield (OP, '=', (eq_r, eq_c), (fs.row, fs.col), fs.src_line(eq_r))
            continue

        # --- Record token start position ---
        tok_r, tok_c = fs.row, fs.col

        # --- Triple-quoted string / f-string ---
        if raw_tok in triple_quoted:
            # PseudoToken matched only the PREFIX+QUOTES; find the body+end
            prefix, quote = _fstring_info(raw_tok)
            endprog_key = raw_tok
            endprog = _compile(endpats.get(endprog_key, Double3))
            fs.advance_by(raw_tok)
            endm = endprog.match(fs.body, fs.pos)
            if endm:
                body_and_close = fs.body[fs.pos:endm.end(0)]
                full_tok = raw_tok + body_and_close
                fs.advance_by(body_and_close)
                if prefix is not None:
                    yield from _fstring_tokens(full_tok, tok_r, tok_c,
                                               fs.source_lines)
                else:
                    yield (STRING, full_tok, (tok_r, tok_c),
                           (fs.row, fs.col), fs.src_line(tok_r))
            else:
                # Unterminated triple-quoted string: just yield what we have
                yield (ERRORTOKEN, raw_tok, (tok_r, tok_c),
                       (fs.row, fs.col), fs.src_line(tok_r))
            continue

        # --- Single-quoted string / f-string ---
        if (initial in single_quoted or
                raw_tok[:2] in single_quoted or
                raw_tok[:3] in single_quoted):
            prefix, quote = _fstring_info(raw_tok)
            fs.advance_by(raw_tok)
            if prefix is not None:
                yield from _fstring_tokens(raw_tok, tok_r, tok_c,
                                           fs.source_lines)
            else:
                yield (STRING, raw_tok, (tok_r, tok_c),
                       (fs.row, fs.col), fs.src_line(tok_r))
            continue

        # --- Newline (inside paren) ---
        if initial in '\r\n':
            fs.advance_by(raw_tok)
            yield (NL, raw_tok, (tok_r, tok_c), (fs.row, fs.col),
                   fs.src_line(tok_r))
            continue

        # --- Comment ---
        if initial == '#':
            fs.advance_by(raw_tok)
            # Comments inside f-string expressions are syntax errors in 3.12,
            # but we emit them for robustness.
            yield (COMMENT, raw_tok, (tok_r, tok_c), (fs.row, fs.col),
                   fs.src_line(tok_r))
            continue

        # --- Identifier ---
        if raw_tok.replace('_', 'a').isidentifier():
            fs.advance_by(raw_tok)
            yield (NAME, raw_tok, (tok_r, tok_c), (fs.row, fs.col),
                   fs.src_line(tok_r))
            continue

        # --- Number ---
        if initial in '0123456789' or (initial == '.' and raw_tok not in ('.', '...')):
            fs.advance_by(raw_tok)
            yield (NUMBER, raw_tok, (tok_r, tok_c), (fs.row, fs.col),
                   fs.src_line(tok_r))
            continue

        # --- Punctuation / operators ---
        fs.advance_by(raw_tok)
        if initial in '([{':
            paren_depth += 1
        elif initial in ')]}':
            paren_depth -= 1
        yield (OP, raw_tok, (tok_r, tok_c), (fs.row, fs.col),
               fs.src_line(tok_r))


def _scan_format_spec(fs):
    """
    Yield tokens for a format spec (after :), up to and including the }.
    The format spec content is emitted as FSTRING_MIDDLE.
    Nested {expr} inside are tokenized recursively via _scan_expr.
    """
    seg_row, seg_col = fs.row, fs.col
    seg = []

    while not fs.at_end():
        ch = fs.peek()

        if ch == '{':
            if fs.peek(1) == '{':
                seg.append('{')
                fs.advance()
                fs.advance()
            else:
                if seg:
                    yield (FSTRING_MIDDLE, ''.join(seg),
                           (seg_row, seg_col), (fs.row, fs.col),
                           fs.src_line(seg_row))
                    seg = []
                open_r, open_c = fs.row, fs.col
                fs.advance()
                yield (OP, '{', (open_r, open_c), (fs.row, fs.col),
                       fs.src_line(open_r))
                yield from _scan_expr(fs)
                seg_row, seg_col = fs.row, fs.col
        elif ch == '}':
            if fs.peek(1) == '}':
                seg.append('}')
                fs.advance()
                fs.advance()
            else:
                if seg:
                    yield (FSTRING_MIDDLE, ''.join(seg),
                           (seg_row, seg_col), (fs.row, fs.col),
                           fs.src_line(seg_row))
                    seg = []
                close_r, close_c = fs.row, fs.col
                fs.advance()
                yield (OP, '}', (close_r, close_c), (fs.row, fs.col),
                       fs.src_line(close_r))
                return
        else:
            seg.append(ch)
            fs.advance()


def _fstring_tokens(token, srow, scol, source_lines):
    """
    Expand a complete f-string token into FSTRING_START / body / FSTRING_END.

    token: complete f-string text (prefix + quotes + body + quotes)
    srow, scol: 1-based row, 0-based column of the first char
    source_lines: list of source lines (for the 'line' field of each token)
    """
    prefix, quote = _fstring_info(token)
    if prefix is None:
        # Shouldn't happen, but fall back to STRING
        sl = source_lines[srow - 1] if source_lines and srow <= len(source_lines) else ''
        yield (STRING, token, (srow, scol), (srow, scol + len(token)), sl)
        return

    fstart = prefix + quote
    body_offset = len(fstart)
    body = token[body_offset:len(token) - len(quote)]

    sl0 = source_lines[srow - 1] if source_lines and srow <= len(source_lines) else ''
    yield (FSTRING_START, fstart,
           (srow, scol), (srow, scol + body_offset), sl0)

    fs = _FStr(body, srow, scol + body_offset, source_lines)
    yield from _scan_body(fs)

    end_r, end_c = fs.row, fs.col
    sl_end = source_lines[end_r - 1] if source_lines and end_r <= len(source_lines) else ''
    yield (FSTRING_END, quote,
           (end_r, end_c), (end_r, end_c + len(quote)), sl_end)


# -- Main tokenizer generator -------------------------------------------------

def _generate(readline, encoding, extra_tokens):
    """
    Tokenize Python source.  Adapted from the Python 3.11 pure-Python
    tokenizer with f-string expansion for 3.12 compatibility.
    """
    lnum = parenlev = continued = 0
    contstr = ''
    needcont = False
    contline = None
    indents = [0]
    source_lines = []   # accumulated for f-string position info

    # Typing hint: line is str after decoding
    last_line = ''
    line = ''
    endprog = _compile(Single)  # placeholder
    strstart = (0, 0)

    while True:
        try:
            last_line = line
            raw = readline()
        except StopIteration:
            raw = b'' if encoding else ''

        if encoding is not None:
            # bytes mode (tokenize): readline must yield bytes
            if not isinstance(raw, (bytes, bytearray)):
                raise TypeError("readline() returned a non-bytes object")
            line = raw.decode(encoding)
        else:
            # str mode (generate_tokens): readline must yield str
            if not isinstance(raw, str):
                raise TypeError("readline() returned a non-string object")
            line = raw

        source_lines.append(line)
        lnum += 1
        pos, maxpos = 0, len(line)

        if contstr:                                    # continued string
            if not line:
                raise SyntaxError("EOF in multi-line string", (None, strstart))
            endmatch = endprog.match(line)
            if endmatch:
                pos = end = endmatch.end(0)
                full_str = contstr + line[:end]
                prefix, _ = _fstring_info(full_str)
                if prefix is not None:
                    yield from _fstring_tokens(full_str, strstart[0], strstart[1],
                                               source_lines)
                else:
                    yield (STRING, full_str, strstart, (lnum, end),
                           contline + line)
                contstr = ''
                needcont = False
                contline = None
            elif needcont and line[-2:] != '\\\n' and line[-3:] != '\\\r\n':
                yield (ERRORTOKEN, contstr + line, strstart, (lnum, len(line)),
                       contline)
                contstr = ''
                contline = None
                continue
            else:
                contstr = contstr + line
                contline = contline + line
                continue

        elif parenlev == 0 and not continued:          # new statement
            if not line:
                break
            column = 0
            while pos < maxpos:
                c = line[pos]
                if c == ' ':
                    column += 1
                elif c == '\t':
                    column = (column // tabsize + 1) * tabsize
                elif c == '\f':
                    column = 0
                else:
                    break
                pos += 1
            if pos == maxpos:
                # whitespace-only final line without a trailing newline: the C
                # tokenizer still emits a width-1 virtual NL before ENDMARKER
                if extra_tokens and line:
                    yield (NL, '', (lnum, pos), (lnum, pos + 1), line)
                break

            if line[pos] in '#\r\n':
                if line[pos] == '#':
                    comment_token = line[pos:].rstrip('\r\n')
                    if extra_tokens:
                        yield (COMMENT, comment_token,
                               (lnum, pos), (lnum, pos + len(comment_token)), line)
                    pos += len(comment_token)
                if extra_tokens:
                    # a missing trailing newline still counts as a width-1 NL
                    nl_end = len(line) + (0 if line.endswith(('\n', '\r')) else 1)
                    yield (NL, line[pos:],
                           (lnum, pos), (lnum, nl_end), line)
                continue

            if column > indents[-1]:
                indents.append(column)
                yield (INDENT, line[:pos], (lnum, 0), (lnum, pos), line)
            while column < indents[-1]:
                if column not in indents:
                    stripped = line.rstrip('\r\n')
                    raise IndentationError(
                        "unindent does not match any outer indentation level",
                        ("<string>", lnum, len(stripped) + 1, stripped))
                indents = indents[:-1]
                yield (DEDENT, '', (lnum, pos), (lnum, pos), line)

        else:                                          # continued statement
            if not line:
                raise SyntaxError("EOF in multi-line statement", (lnum, 0))
            continued = 0

        while pos < maxpos:
            pseudomatch = _compile(PseudoToken).match(line, pos)
            if pseudomatch:
                start, end = pseudomatch.span(1)
                spos, epos, pos = (lnum, start), (lnum, end), end
                if start == end:
                    continue
                token, initial = line[start:end], line[start]

                if (initial in '0123456789' or
                        (initial == '.' and token not in ('.', '...'))):
                    yield (NUMBER, token, spos, epos, line)

                elif initial in '\r\n':
                    if parenlev > 0:
                        if extra_tokens:
                            yield (NL, token, spos, epos, line)
                    else:
                        yield (NEWLINE, token, spos, epos, line)

                elif initial == '#':
                    if extra_tokens:
                        yield (COMMENT, token, spos, epos, line)

                elif token in triple_quoted:
                    endprog = _compile(endpats[token])
                    endmatch = endprog.match(line, pos)
                    if endmatch:
                        pos = endmatch.end(0)
                        token = line[start:pos]
                        prefix, _ = _fstring_info(token)
                        if prefix is not None:
                            yield from _fstring_tokens(token, lnum, start,
                                                       source_lines)
                        else:
                            yield (STRING, token, spos, (lnum, pos), line)
                    else:
                        strstart = (lnum, start)
                        contstr = line[start:]
                        contline = line
                        break

                elif (initial in single_quoted or
                      token[:2] in single_quoted or
                      token[:3] in single_quoted):
                    if token[-1] == '\n':              # continued string
                        strstart = (lnum, start)
                        endprog = _compile(
                            endpats.get(initial) or
                            endpats.get(token[1]) or
                            endpats.get(token[2]))
                        contstr = line[start:]
                        needcont = True
                        contline = line
                        break
                    else:
                        prefix, _ = _fstring_info(token)
                        if prefix is not None:
                            yield from _fstring_tokens(token, lnum, start,
                                                       source_lines)
                        else:
                            yield (STRING, token, spos, epos, line)

                elif initial.isidentifier():
                    yield (NAME, token, spos, epos, line)

                elif initial == '\\':
                    continued = 1

                else:
                    if initial in '([{':
                        parenlev += 1
                    elif initial in ')]}':
                        # don't go negative on an unmatched close bracket, or
                        # EOF gets misread as an unterminated continued statement
                        if parenlev > 0:
                            parenlev -= 1
                    yield (OP, token, spos, epos, line)
            else:
                yield (ERRORTOKEN, line[pos], (lnum, pos), (lnum, pos + 1), line)
                pos += 1

    # Implicit NEWLINE if the last line didn't end with one
    if last_line and last_line[-1] not in '\r\n':
        if not last_line.strip().startswith('#'):
            last_lnum = lnum - 1
            yield (NEWLINE, '',
                   (last_lnum, len(last_line)),
                   (last_lnum, len(last_line) + 1), last_line)

    for _ in indents[1:]:
        yield (DEDENT, '', (lnum, 0), (lnum, 0), '')

    # ENDMARKER sits on the line after the last content line.  When we broke on
    # an empty readline (EOF), lnum already counts that empty line; when we broke
    # on a content line (e.g. whitespace-only, no trailing newline), it doesn't.
    endmarker_lnum = lnum
    if extra_tokens and line != '':
        endmarker_lnum += 1
    yield (ENDMARKER, '', (endmarker_lnum, 0), (endmarker_lnum, 0), '')


# -- Public API ---------------------------------------------------------------

class TokenizerIter:
    """
    Iterator that tokenizes Python source code, matching CPython 3.12's
    _tokenize.TokenizerIter interface.

    TokenizerIter(readline, *, extra_tokens=False, encoding='utf-8')

    readline must be a callable returning one line at a time (str or bytes).
    Each iteration yields a 5-tuple:
        (type, string, (start_row, start_col), (end_row, end_col), line)
    """

    def __init__(self, readline, *, extra_tokens=False, encoding=None):
        # encoding=None means str mode (readline yields str, no decoding);
        # an explicit encoding means bytes mode.  This mirrors CPython, where
        # tokenize() passes encoding and generate_tokens() does not.
        self._iter = _generate(readline, encoding, extra_tokens)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iter)



import unittest
import sys
from test import support
from ctypes import *

class StructFieldsTestCase(unittest.TestCase):
    # Structure/Union classes must get 'finalized' sooner or
    # later, when one of these things happen:
    #
    # 1. _fields_ is set.
    # 2. An instance is created.
    # 3. The type is used as field of another Structure/Union.
    # 4. The type is subclassed
    #
    # When they are finalized, assigning _fields_ is no longer allowed.

    def test_1_A(self):
        class X(Structure):
            pass
        self.assertEqual(sizeof(X), 0) # not finalized
        X._fields_ = [] # finalized
        self.assertRaises(AttributeError, setattr, X, "_fields_", [])

    def test_1_B(self):
        class X(Structure):
            _fields_ = [] # finalized
        self.assertRaises(AttributeError, setattr, X, "_fields_", [])

    def test_2(self):
        class X(Structure):
            pass
        X()
        self.assertRaises(AttributeError, setattr, X, "_fields_", [])

    def test_3(self):
        class X(Structure):
            pass
        class Y(Structure):
            _fields_ = [("x", X)] # finalizes X
        self.assertRaises(AttributeError, setattr, X, "_fields_", [])

    def test_4(self):
        class X(Structure):
            pass
        class Y(X):
            pass
        self.assertRaises(AttributeError, setattr, X, "_fields_", [])
        Y._fields_ = []
        self.assertRaises(AttributeError, setattr, X, "_fields_", [])

    @support.cpython_only
    def test_5(self):
        class X(Structure):
            _fields_ = (("char", c_char * 5),)

        x = X(b'#' * 5)
        x.char = b'a\0b\0'
        self.assertEqual(bytes(x), b'a\x00###')

    def test_6(self):
        class X(Structure):
            _fields_ = [("x", c_int)]
        CField = type(X.x)
        self.assertRaises(TypeError, CField)

    @support.cpython_only
    def test_gh99275(self):
        class BrokenStructure(Structure):
            def __init_subclass__(cls, **kwargs):
                cls._fields_ = []  # This line will fail, `stgdict` is not ready

        with self.assertRaisesRegex(TypeError,
                                    'ctypes state is not initialized'):
            class Subclass(BrokenStructure): ...

    def test_max_field_size_gh126937(self):
        # Classes for big structs should be created successfully.
        # (But they most likely can't be instantiated.)
        # The size must fit in Py_ssize_t.

        class X(Structure):
            _fields_ = [('char', c_char),]
        max_field_size = sys.maxsize

        class Y(Structure):
            _fields_ = [('largeField', X * max_field_size)]
        class Z(Structure):
            _fields_ = [('largeField', c_char * max_field_size)]

        with self.assertRaises(OverflowError):
            class TooBig(Structure):
                _fields_ = [('largeField', X * (max_field_size + 1))]
        with self.assertRaises(OverflowError):
            class TooBig(Structure):
                _fields_ = [('largeField', c_char * (max_field_size + 1))]

    # __set__ and __get__ should raise a TypeError in case their self
    # argument is not a ctype instance.
    def test___set__(self):
        class MyCStruct(Structure):
            _fields_ = (("field", c_int),)
        self.assertRaises(TypeError,
                          MyCStruct.field.__set__, 'wrong type self', 42)

        class MyCUnion(Union):
            _fields_ = (("field", c_int),)
        self.assertRaises(TypeError,
                          MyCUnion.field.__set__, 'wrong type self', 42)

    def test___get__(self):
        class MyCStruct(Structure):
            _fields_ = (("field", c_int),)
        self.assertRaises(TypeError,
                          MyCStruct.field.__get__, 'wrong type self', 42)

        class MyCUnion(Union):
            _fields_ = (("field", c_int),)
        self.assertRaises(TypeError,
                          MyCUnion.field.__get__, 'wrong type self', 42)

if __name__ == "__main__":
    unittest.main()



import os
import sys
from _cffi_ssl import _cffi_src
sys.modules['_cffi_src'] = _cffi_src
#
from _cffi_ssl._cffi_src.build_openssl import (build_ffi_for_binding,
        _get_openssl_libraries, extra_link_args, compiler_type)

if sys.platform == "win32":
    pypy_win32_extra = ["pypy_win32_extra"]
else:
    pypy_win32_extra = []

libraries=_get_openssl_libraries(sys.platform)

# openssl headers/libraries frequently live outside the compiler's default
# search path (Homebrew's keg-only openssl@3, a custom prefix on CI, ...).
# Point the build at them by exporting CFLAGS/LDFLAGS, e.g. on macOS:
#     CFLAGS="-I$(brew --prefix openssl@3)/include" \
#     LDFLAGS="-L$(brew --prefix openssl@3)/lib" \
#         pypy build_cffi_imports.py
# The compiler driver already forwards CFLAGS/CPPFLAGS/LDFLAGS from the
# environment, so we do not hard-code any platform path here; /usr/local is
# only added as a fallback when it actually exists (Intel Homebrew, the
# benchmarker).
extra_compile_args = []
if sys.platform != 'win32' and os.path.isdir('/usr/local/include'):
    extra_compile_args.append('-I/usr/local/include')

ffi = build_ffi_for_binding(
    module_name="_pypy_openssl",
    module_prefix="_cffi_src.openssl.",
    modules=[
        # This goes first so we can define some cryptography-wide symbols.
        "cryptography",

        # Provider comes early as well so we define OSSL_LIB_CTX
        "provider",
        "asn1",
        "bignum",
        "bio",
        "cmac",
        "crypto",
        "ct",
        "dh",
        "dsa",
        "ec",
        "ecdh",
        "ecdsa",
        "engine",
        "err",
        "evp",
        "fips",
        "hmac",
        "nid",
        "objects",
        "ocsp",
        "opensslv",
        "pem",
        "pkcs12",
        "rand",
        "rsa",
        "ssl",
        "x509",
        "x509name",
        "x509v3",
        "x509_vfy",
        "pkcs7",
        "callbacks",
    ] + pypy_win32_extra,
    libraries=libraries,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args(compiler_type()),
)

if __name__ == '__main__':
    ffi.compile(verbose=True)
    if sys.platform == 'win32' and "PYPY_PACKAGE_NO_DLLS" not in os.environ:
        # copy dlls from externals to the pwd
        # maybe we should link to libraries instead of the dlls
        # to avoid this mess
        import os, glob, shutil
        path_parts = os.environ['PATH'].split(';')
        candidates = [x for x in path_parts if 'externals' in x]

        def copy_from_path(dll):
            for c in candidates:
                files = glob.glob(os.path.join(c, dll + '*.dll'))
                if files:
                    for fname in files:
                        print('copying', fname)
                        shutil.copy(fname, '.')
                    break
                else:
                    print("not copying %s from %s", (dll, c))

        if candidates:
            for lib in libraries:
                copy_from_path(lib)
        else:
            print('no "externals" on PATH, not copying %s, expect trouble', libraries)



###############################################################################
# Server process to keep track of unlinked resources (like shared memory
# segments, semaphores etc.) and clean them.
#
# On Unix we run a server process which keeps track of unlinked
# resources. The server ignores SIGINT and SIGTERM and reads from a
# pipe.  Every other process of the program has a copy of the writable
# end of the pipe, so we get EOF when all other processes have exited.
# Then the server process unlinks any remaining resource names.
#
# This is important because there may be system limits for such resources: for
# instance, the system only supports a limited number of named semaphores, and
# shared-memory segments live in the RAM. If a python process leaks such a
# resource, this resource will not be removed till the next reboot.  Without
# this resource tracker process, "killall python" would probably leave unlinked
# resources.

import os
import signal
import sys
import threading
import warnings

from . import spawn
from . import util

__all__ = ['ensure_running', 'register', 'unregister']

_HAVE_SIGMASK = hasattr(signal, 'pthread_sigmask')
_IGNORED_SIGNALS = (signal.SIGINT, signal.SIGTERM)

_CLEANUP_FUNCS = {
    'noop': lambda: None,
}

if os.name == 'posix':
    import _multiprocessing
    import _posixshmem

    # Use sem_unlink() to clean up named semaphores.
    #
    # sem_unlink() may be missing if the Python build process detected the
    # absence of POSIX named semaphores. In that case, no named semaphores were
    # ever opened, so no cleanup would be necessary.
    if hasattr(_multiprocessing, 'sem_unlink'):
        _CLEANUP_FUNCS.update({
            'semaphore': _multiprocessing.sem_unlink,
        })
    _CLEANUP_FUNCS.update({
        'shared_memory': _posixshmem.shm_unlink,
    })


class ReentrantCallError(RuntimeError):
    pass


class ResourceTracker(object):

    def __init__(self):
        self._lock = threading.RLock()
        self._fd = None
        self._pid = None
        # PYPY change: shadow registry for restart resilience -- if the tracker
        # subprocess dies and is relaunched, we replay all outstanding REGISTERs
        # so the new tracker has a complete view of resources to clean up.
        self._registered = {}
        # PYPY change end

    def _reentrant_call_error(self):
        # gh-109629: this happens if an explicit call to the ResourceTracker
        # gets interrupted by a garbage collection, invoking a finalizer (*)
        # that itself calls back into ResourceTracker.
        #   (*) for example the SemLock finalizer
        raise ReentrantCallError(
            "Reentrant call into the multiprocessing resource tracker")

    def __del__(self):
        # making sure child processess are cleaned before ResourceTracker
        # gets destructed.
        # see https://github.com/python/cpython/issues/88887
        self._stop(use_blocking_lock=False)

    def _stop(self, use_blocking_lock=True):
        if use_blocking_lock:
            with self._lock:
                self._stop_locked()
        else:
            acquired = self._lock.acquire(blocking=False)
            try:
                self._stop_locked()
            finally:
                if acquired:
                    self._lock.release()

    def _stop_locked(
        self,
        close=os.close,
        waitpid=os.waitpid,
        waitstatus_to_exitcode=os.waitstatus_to_exitcode,
    ):
        # This shouldn't happen (it might when called by a finalizer)
        # so we check for it anyway.
        if self._lock._recursion_count() > 1:
            return self._reentrant_call_error()
        if self._fd is None:
            # not running
            return
        if self._pid is None:
            return

        # closing the "alive" file descriptor stops main()
        close(self._fd)
        self._fd = None

        waitpid(self._pid, 0)
        self._pid = None

    def getfd(self):
        self.ensure_running()
        return self._fd

    def ensure_running(self):
        '''Make sure that resource tracker process is running.

        This can be run from any process.  Usually a child process will use
        the resource created by its parent.'''
        with self._lock:
            if self._lock._recursion_count() > 1:
                # The code below is certainly not reentrant-safe, so bail out
                return self._reentrant_call_error()
            if self._fd is not None:
                # resource tracker was launched before, is it still running?
                if self._check_alive():
                    # => still alive
                    return
                # => dead, launch it again
                os.close(self._fd)

                # Clean-up to avoid dangling processes.
                try:
                    # _pid can be None if this process is a child from another
                    # python process, which has started the resource_tracker.
                    if self._pid is not None:
                        os.waitpid(self._pid, 0)
                except ChildProcessError:
                    # The resource_tracker has already been terminated.
                    pass
                self._fd = None
                self._pid = None

                warnings.warn('resource_tracker: process died unexpectedly, '
                              'relaunching.  Some resources might leak.')

            fds_to_pass = []
            try:
                fds_to_pass.append(sys.stderr.fileno())
            except Exception:
                pass
            cmd = 'from multiprocessing.resource_tracker import main;main(%d)'
            r, w = os.pipe()
            try:
                fds_to_pass.append(r)
                # process will out live us, so no need to wait on pid
                exe = spawn.get_executable()
                args = [exe] + util._args_from_interpreter_flags()
                args += ['-c', cmd % r]
                # bpo-33613: Register a signal mask that will block the signals.
                # This signal mask will be inherited by the child that is going
                # to be spawned and will protect the child from a race condition
                # that can make the child die before it registers signal handlers
                # for SIGINT and SIGTERM. The mask is unregistered after spawning
                # the child.
                prev_sigmask = None
                try:
                    if _HAVE_SIGMASK:
                        prev_sigmask = signal.pthread_sigmask(signal.SIG_BLOCK, _IGNORED_SIGNALS)
                    pid = util.spawnv_passfds(exe, args, fds_to_pass)
                finally:
                    if prev_sigmask is not None:
                        signal.pthread_sigmask(signal.SIG_SETMASK, prev_sigmask)
            except:
                os.close(w)
                raise
            else:
                self._fd = w
                self._pid = pid
                # PYPY change: replay outstanding registrations to the new
                # tracker.  On first start _registered is empty so this is a
                # no-op; on restart it re-syncs the new tracker's cache.
                for pending_name, pending_rtype in list(self._registered.items()):
                    replay_msg = 'REGISTER:{0}:{1}\n'.format(
                        pending_name, pending_rtype).encode('ascii')
                    os.write(self._fd, replay_msg)
                # PYPY change end
            finally:
                os.close(r)

    def _check_alive(self):
        '''Check that the pipe has not been closed by sending a probe.'''
        try:
            # We cannot use send here as it calls ensure_running, creating
            # a cycle.
            os.write(self._fd, b'PROBE:0:noop\n')
        except OSError:
            return False
        else:
            return True

    def register(self, name, rtype):
        '''Register name of resource with resource tracker.'''
        # PYPY change: send BEFORE updating the shadow registry so that if
        # _send raises (e.g. ValueError for a too-long name) the name is not
        # left as a stale entry that would be replayed to a restarted tracker.
        # This ordering is safe: ensure_running() inside _send replays
        # _registered before os.write() is called, so adding the name
        # afterwards still lands in _registered in time for any subsequent
        # restart.
        self._send('REGISTER', name, rtype)
        self._registered[name] = rtype
        # PYPY change end

    def unregister(self, name, rtype):
        '''Unregister name of resource with resource tracker.'''
        self._send('UNREGISTER', name, rtype)
        # PYPY change: remove from shadow registry *after* sending so that a
        # tracker restart during send still replays the entry before the
        # UNREGISTER lands on the new tracker.
        self._registered.pop(name, None)
        # PYPY change end

    def _send(self, cmd, name, rtype):
        try:
            self.ensure_running()
        except ReentrantCallError:
            # The code below might or might not work, depending on whether
            # the resource tracker was already running and still alive.
            # Better warn the user.
            # (XXX is warnings.warn itself reentrant-safe? :-)
            warnings.warn(
                f"ResourceTracker called reentrantly for resource cleanup, "
                f"which is unsupported. "
                f"The {rtype} object {name!r} might leak.")
        # PYPY change: only bail out when _fd is None, i.e. we are in the
        # middle of (re)starting the tracker and os.write(None,...) would
        # crash.  When _fd is valid the tracker is alive and the write
        # below is safe even though ensure_running() detected reentrancy;
        # delivering the message avoids a spurious "leaked / ENOENT" pair
        # at shutdown caused by the tracker never receiving the UNREGISTER.
        if self._fd is None:
             return
        # PYPY change end
        msg = '{0}:{1}:{2}\n'.format(cmd, name, rtype).encode('ascii')
        if len(msg) > 512:
            # posix guarantees that writes to a pipe of less than PIPE_BUF
            # bytes are atomic, and that PIPE_BUF >= 512
            raise ValueError('msg too long')
        nbytes = os.write(self._fd, msg)
        assert nbytes == len(msg), "nbytes {0:n} but len(msg) {1:n}".format(
            nbytes, len(msg))


_resource_tracker = ResourceTracker()
ensure_running = _resource_tracker.ensure_running
register = _resource_tracker.register
unregister = _resource_tracker.unregister
getfd = _resource_tracker.getfd


def main(fd):
    '''Run resource tracker.'''
    # protect the process from ^C and "killall python" etc
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)
    if _HAVE_SIGMASK:
        signal.pthread_sigmask(signal.SIG_UNBLOCK, _IGNORED_SIGNALS)

    for f in (sys.stdin, sys.stdout):
        try:
            f.close()
        except Exception:
            pass

    cache = {rtype: set() for rtype in _CLEANUP_FUNCS.keys()}
    try:
        # keep track of registered/unregistered resources
        with open(fd, 'rb') as f:
            for line in f:
                try:
                    cmd, name, rtype = line.strip().decode('ascii').split(':')
                    cleanup_func = _CLEANUP_FUNCS.get(rtype, None)
                    if cleanup_func is None:
                        raise ValueError(
                            f'Cannot register {name} for automatic cleanup: '
                            f'unknown resource type {rtype}')

                    if cmd == 'REGISTER':
                        cache[rtype].add(name)
                    elif cmd == 'UNREGISTER':
                        # PYPY change: use discard instead of remove so an
                        # out-of-order UNREGISTER (arriving before its REGISTER
                        # due to a tracker restart or reentrant GC) does not
                        # print a KeyError traceback to stderr and fail tests.
                        cache[rtype].discard(name)
                        # PYPY change end
                    elif cmd == 'PROBE':
                        pass
                    else:
                        raise RuntimeError('unrecognized command %r' % cmd)
                except Exception:
                    try:
                        sys.excepthook(*sys.exc_info())
                    except:
                        pass
    finally:
        # all processes have terminated; cleanup any remaining resources
        for rtype, rtype_cache in cache.items():
            if rtype_cache:
                try:
                    warnings.warn('resource_tracker: There appear to be %d '
                                  'leaked %s objects to clean up at shutdown' %
                                  (len(rtype_cache), rtype))
                except Exception:
                    pass
            for name in rtype_cache:
                # For some reason the process which created and registered this
                # resource has failed to unregister it. Presumably it has
                # died.  We therefore unlink it.
                try:
                    try:
                        _CLEANUP_FUNCS[rtype](name)
                    except Exception as e:
                        warnings.warn('resource_tracker: %r: %s' % (name, e))
                finally:
                    pass
